from __future__ import annotations

from typing import (
    List,
    Literal,
    TYPE_CHECKING,
    Union,
)

from pyvenafi.tpp.api.api_base import InvalidResponseError
from pyvenafi.tpp.attributes.discovery import DiscoveryAttributes
from pyvenafi.tpp.attributes.onboard_discovery import OnboardDiscoveryAttributes
from pyvenafi.tpp.features.bases.feature_base import (
    feature,
    FeatureBase,
)
from pyvenafi.tpp.features.definitions.exceptions import UnexpectedValue

if TYPE_CHECKING:
    from pyvenafi.tpp.api.websdk.models import (
        config,
        identity as ident,
    )

@feature('Network Discovery')
class NetworkDiscovery(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        hosts: list[str],
        default_certificate_location: 'Union[config.Object, str]',
        attributes: dict = None,
        automatically_import: bool = False,
        blackout: dict[str, List] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        exclusion_locations: 'Union[config.Object, str]' = None,
        hour: int = None,
        placement_rules: 'list[Union[config.Object, str]]' = None,
        ports: list[Union[str, int]] = None,
        priority: int = None,
        reschedule: bool = True,
        resolve_host: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates a Network Discovery job.

        Args:
            name: Name of the discovery job.
            hosts: A list of hosts. If specific hosts should scan different ports, then specify by appending the
                   port to the IP address or hostname (i.e. 192.168.0.10:80). If no port is specified, then the
                   `ports` parameter will be appended to those hosts.
            default_certificate_location: :ref:`config_object` or :ref:`dn` of the folder to place results not 
                                          matching a placement rule.
            attributes: Additional attributes.
            automatically_import: If `True`, the job will terminate by placing the certificate objects, device
                                  objects, and application objects according to the placement rules.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            exclusion_locations: List of :ref:`config_object` or :ref:`dn` of exclusion folders.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            placement_rules: List of :ref:`config_object` or :ref:`dn` for the placement rules. The order of the list matters
                             as the rules are prioritized accordingly.
            ports: List of ports to scan.
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            resolve_host: Resolve the hostname when ``True``.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        placement_rule_guids = None
        if placement_rules:
            placement_rule_guids = [f'{e}:{self._get_guid(pr)}' for e, pr in enumerate(placement_rules)]
        ports = ','.join(list(map(str, ports))) if ports else \
            "22,80,443-449,465,563,636,695,981-995,1311,1920,2083,2087,2096,2211,2484,2949,3268,3269,3414," \
            "4712,4843,5223,5358,6619,6679,6697,7002,8080,8222,8243,8333,8443,8878,8881,8882,9043,9090," \
            "9091,9443,18072,18080-18085,18090-18094,28080"
        address_range = [f'{h}:{ports}' if ':' not in h else h for h in hosts]

        if blackout:
            blackout = [f'{k}:{",".join(v)}' for k, v in blackout.items()]

        attributes = attributes or {}
        attributes.update(
            {
                DiscoveryAttributes.address_range          : address_range,
                DiscoveryAttributes.automatically_import   : "1" if automatically_import else "0",
                DiscoveryAttributes.blackout               : blackout,
                DiscoveryAttributes.certificate_location_dn: self._get_dn(default_certificate_location)
                if default_certificate_location else None,
                DiscoveryAttributes.contact                : [self._get_prefixed_universal(c) for c in
                                                              contacts] if contacts else None,
                DiscoveryAttributes.description            : description,
                DiscoveryAttributes.discovery_exclusion_dn : [self._get_dn(el) for el in exclusion_locations]
                if exclusion_locations else None,
                DiscoveryAttributes.placement_rule         : placement_rule_guids,
                DiscoveryAttributes.priority               : priority,
                DiscoveryAttributes.reschedule             : "1" if hour and reschedule else "0",
                DiscoveryAttributes.resolve_host           : "1" if resolve_host else "0",
                DiscoveryAttributes.utc                    : utc
            }
        )

        if hour:
            attributes[DiscoveryAttributes.hour] = hour
            if days_of_week:
                attributes[DiscoveryAttributes.days_of_week] = days_of_week
            elif days_of_month:
                attributes[DiscoveryAttributes.days_of_month] = days_of_month
            elif days_of_year:
                attributes[DiscoveryAttributes.days_of_year] = days_of_year

        return self._config_create(
            name=name,
            parent_folder_dn=self._discovery_dn,
            config_class=DiscoveryAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

    def delete(self, job: 'Union[config.Object, str]'):
        """
        Deletes the discovery job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_guid = self._get_guid(job, parent_dn=self._discovery_dn)
        self._api.websdk.Discovery.Guid(guid=job_guid).delete()

    def get(self, name: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: Name of the discovery job.
            raise_error_if_not_exists: Raise an exception if the discovery job does not exist.

        Returns:
            :ref:`config_object` of the discovery job.
        """
        return self._get_config_object(
            object_dn=f'{self._discovery_dn}\\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def is_in_progress(self, job: 'Union[config.Object, str]'):
        """
        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.

        Returns:
            bool: ``True`` if the job is in progress or ``False`` if it is not.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        try:
            response = self._api.websdk.Config.Read.post(
                object_dn=job_dn,
                attribute_name=DiscoveryAttributes.status
            )
            in_progress_states = ['Pending Execution', 'Running']
            if len(response.values) > 0:
                status = response.values[0]
                return status in in_progress_states
            return False
        except InvalidResponseError:
            return False

    def schedule(
        self, job: 'Union[config.Object, str]', hour: Union[str, int], days_of_week: list[Union[int, str]] = None,
        days_of_month: list[Union[int, str]] = None, days_of_year: list[str] = None
    ):
        """
        Schedules an existing job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            hour: 24-hour UTC hour format (i.e. 20 = 8PM UTC).
            days_of_week: Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Days of the month without leading zeros.
            days_of_year: Days of the year in "MM/DD" format without leading zeros (i.e. 1/23, 10/3).
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        hour = str(hour)
        if not isinstance(hour, list):
            hour = [hour]

        attributes = {
            DiscoveryAttributes.reschedule: "1",
            DiscoveryAttributes.hour      : hour
        }
        if days_of_week:
            attributes[DiscoveryAttributes.days_of_week] = map(str, days_of_week)
        elif days_of_month:
            attributes[DiscoveryAttributes.days_of_month] = map(str, days_of_month)
        elif days_of_year:
            attributes[DiscoveryAttributes.days_of_year] = days_of_year

        self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(attributes)
        )

    def unschedule(self, job: 'Union[config.Object, str]'):
        """
        Removes a schedule from a job. This does not delete the job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        for attribute_name in {
            DiscoveryAttributes.hour,
            DiscoveryAttributes.days_of_year,
            DiscoveryAttributes.days_of_month,
            DiscoveryAttributes.days_of_week,
            DiscoveryAttributes.reschedule
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=job_dn,
                attribute_name=attribute_name
            )

    def blackout_schedule(
        self, job: 'Union[config.Object, str]', sunday: list[Union[str, int]] = None,
        monday: list[Union[str, int]] = None, tuesday: list[Union[str, int]] = None,
        wednesday: list[Union[str, int]] = None, thursday: list[Union[str, int]] = None,
        friday: list[Union[str, int]] = None, saturday: list[Union[str, int]] = None
    ):
        """
        Times of the week to restrict a discovery job from processing.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            sunday: List of hours without leading zeros to restrict processing on Sunday.
            monday: List of hours without leading zeros to restrict processing on Monday.
            tuesday: List of hours without leading zeros to restrict processing on Tuesday.
            wednesday: List of hours without leading zeros to restrict processing on Wednesday.
            thursday: List of hours without leading zeros to restrict processing on Thursday.
            friday: List of hours without leading zeros to restrict processing on Friday.
            saturday: List of hours without leading zeros to restrict processing on Saturday.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        blackout = []
        for e, day in enumerate([sunday, monday, tuesday, wednesday, thursday, friday, saturday]):
            if day:
                hours = ','.join(map(str, day))
                blackout.append(f'{e}:{hours}')
        attributes = {
            DiscoveryAttributes.blackout: blackout
        }
        self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(attributes)
        )

    def run_now(self, job: 'Union[config.Object, str]', timeout: int = 60):
        """
        Runs a job despite any scheduling. This does not return until the job is processing,
        or has a *Processing* Attribute.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            timeout: Timeout in seconds within which the job should start.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(
                {
                    "Start Now": ['1']  # Secret config-bridge attribute name.
                }
            )
        )

        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=0.5):
                if self.is_in_progress(job=job):
                    return

        raise UnexpectedValue(
            f'Expected the job "{job_dn}" to start progress, but it did not.'
        )

    def cancel(self, job: 'Union[config.Object, str]'):
        """
        Cancels a currently running job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Canceled']
        )

    def pause(self, job: 'Union[config.Object, str]'):
        """
        Pauses a currently running job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Paused']
        )

    def resume(self, job: 'Union[config.Object, str]'):
        """
        Resumes a currently paused job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Pending']
        )

    def place_results(self, job: 'Union[config.Object, str]'):
        """
        .. warning::
            This functionality has been deprecated in TPP 21.1 and will have no effect
            from this version forward.

        Places the results of the discovery job according to the placement rules.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        if self._is_version_compatible(maximum="20.4"):
            job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
            self._api.websdk.Config.WriteDn.post(
                object_dn=job_dn,
                attribute_name="Import Results Now",  # Secret config-bridge attribute name.
                values=['1']
            )
        else:
            self._log_warning_message(
                'Cannot run place_results() because it has been deprecated since TPP 21.1. Results '
                'are placed along with discovery beginning in 21.1.'
            )

    def get_all_jobs(self):
        """
        Returns:
            List of :ref:`config_object` of all network discovery jobs.
        """
        jobs = self._api.websdk.Config.FindObjectsOfClass.post(
            class_name='Discovery',
            object_dn=self._discovery_dn
        )

        return jobs.objects

    def wait_for_job_to_finish(self, job: 'Union[config.Object, str]', check_interval: int = 5, timeout: int = 300):
        """
        Waits for the  *Status* attribute to have a value other than *Pending Execution* and *Running*
        on the discovery job. An error is raised if the timeout is exceeded.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            check_interval: Poll interval in seconds to validate that the job finished.
            timeout: Timeout in seconds to wait for the job to finish.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=check_interval):
                if not self.is_in_progress(job=job):
                    return

        status = self._api.websdk.Config.Read.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status
        ).values[0]
        raise TimeoutError(
            f'Expected Network Discovery Job "{job_dn}" to finish within {timeout} seconds, but it is still '
            f'running. It has a status of "{status}."'
        )

class OnboardDiscoveryBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def _create(
        self,
        name: str,
        installation_type: str,
        driver_name: str,
        attributes: dict,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        if blackout:
            blackout = [f'{k}:{",".join(v)}' for k, v in blackout.items()]
        attributes.update(
            {
                OnboardDiscoveryAttributes.application_type             : installation_type,
                OnboardDiscoveryAttributes.driver_name                  : driver_name,
                OnboardDiscoveryAttributes.certificates_placement_folder: self._get_dn(certificate_placement_folder),
                DiscoveryAttributes.blackout                            : blackout,
                DiscoveryAttributes.contact                             : [self._get_prefixed_universal(c) for c in
                                                                           contacts] if contacts else None,
                DiscoveryAttributes.description                         : description,
                DiscoveryAttributes.priority                            : priority,
                DiscoveryAttributes.reschedule                          : "1" if hour and reschedule else "0",
                DiscoveryAttributes.utc                                 : utc,

            }
        )

        if hour:
            attributes[DiscoveryAttributes.hour] = hour
            if days_of_week:
                attributes[DiscoveryAttributes.days_of_week] = days_of_week
            elif days_of_month:
                attributes[DiscoveryAttributes.days_of_month] = days_of_month
            elif days_of_year:
                attributes[DiscoveryAttributes.days_of_year] = days_of_year

        return self._config_create(
            name=name, parent_folder_dn=self._discovery_dn,
            config_class=OnboardDiscoveryAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

    def delete(self, job: 'Union[config.Object, str]'):
        """
        Deletes the discovery job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._secret_store_delete(job_dn)
        self._config_delete(job_dn)

    def get(self, name: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: Name of the discovery job.
            raise_error_if_not_exists: Raise an exception if the discovery job does not exist.

        Returns:
            :ref:`config_object` of the discovery job.
        """
        return self._get_config_object(
            object_dn=f'{self._discovery_dn}\\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def is_in_progress(self, job: 'Union[config.Object, str]'):
        """
        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.

        Returns:
            bool: ``True`` if the job is in progress or ``False`` if it is not.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        try:
            response = self._api.websdk.Config.Read.post(
                object_dn=job_dn,
                attribute_name=DiscoveryAttributes.status
            )
            in_progress_states = ['Pending Execution', 'Running']
            if len(response.values) > 0:
                status = response.values[0]
                return status in in_progress_states
            return False
        except InvalidResponseError:
            return False

    def schedule(
        self, job: 'Union[config.Object, str]', hour: Union[str, int], days_of_week: list[Union[int, str]] = None,
        days_of_month: list[Union[int, str]] = None, days_of_year: list[str] = None
    ):
        """
        Schedules an existing job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            hour: 24-hour UTC hour format (i.e. 20 = 8PM UTC).
            days_of_week: Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Days of the month without leading zeros.
            days_of_year: Days of the year in "MM/DD" format without leading zeros (i.e. 1/23, 10/3).
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        hour = str(hour)

        attributes = {
            DiscoveryAttributes.reschedule: "1",
            DiscoveryAttributes.hour      : hour
        }
        if days_of_week:
            attributes[DiscoveryAttributes.days_of_week] = list(map(str, days_of_week))
        elif days_of_month:
            attributes[DiscoveryAttributes.days_of_month] = list(map(str, days_of_month))
        elif days_of_year:
            attributes[DiscoveryAttributes.days_of_year] = list(map(str, days_of_year))

        self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(attributes)
        )

    def unschedule(self, job: 'Union[config.Object, str]'):
        """
        Removes a schedule from a job. This does not delete the job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        for attribute_name in {
            DiscoveryAttributes.hour,
            DiscoveryAttributes.days_of_year,
            DiscoveryAttributes.days_of_month,
            DiscoveryAttributes.days_of_week,
            DiscoveryAttributes.reschedule
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=job_dn,
                attribute_name=attribute_name
            )

    def blackout_schedule(
        self, job: 'Union[config.Object, str]', sunday: list[Union[str, int]] = None,
        monday: list[Union[str, int]] = None, tuesday: list[Union[str, int]] = None,
        wednesday: list[Union[str, int]] = None, thursday: list[Union[str, int]] = None,
        friday: list[Union[str, int]] = None, saturday: list[Union[str, int]] = None
    ):
        """
        Times of the week to restrict a discovery job from processing.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            sunday: List of hours without leading zeros to restrict processing on Sunday.
            monday: List of hours without leading zeros to restrict processing on Monday.
            tuesday: List of hours without leading zeros to restrict processing on Tuesday.
            wednesday: List of hours without leading zeros to restrict processing on Wednesday.
            thursday: List of hours without leading zeros to restrict processing on Thursday.
            friday: List of hours without leading zeros to restrict processing on Friday.
            saturday: List of hours without leading zeros to restrict processing on Saturday.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        blackout = []
        for e, day in enumerate([sunday, monday, tuesday, wednesday, thursday, friday, saturday]):
            if day:
                hours = ','.join(map(str, day))
                blackout.append(f'{e}:{hours}')
        attributes = {
            DiscoveryAttributes.blackout: blackout
        }
        self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(attributes)
        )

    def run_now(self, job: 'Union[config.Object, str]', timeout: int = 60):
        """
        Runs a job despite any scheduling. This does not return until the job is processing,
        or has a *Processing* Attribute.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            timeout: Timeout in seconds within which the job should start.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(
                {
                    "Start Now": ['1']  # Secret config-bridge attribute name.
                }
            )
        )

        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=0.5):
                if self.is_in_progress(job=job):
                    return

        raise UnexpectedValue(
            f'Expected the job "{job_dn}" to start progress, but it did not.'
        )

    def cancel(self, job: 'Union[config.Object, str]'):
        """
        Cancels a currently running job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Canceled']
        )

    def pause(self, job: 'Union[config.Object, str]'):
        """
        Pauses a currently running job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Paused']
        )

    def resume(self, job: 'Union[config.Object, str]'):
        """
        Resumes a currently paused job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Pending']
        )

    def get_all_jobs(self):
        """
        Returns:
            List of :ref:`config_object` of all network discovery jobs.
        """
        jobs = self._api.websdk.Config.FindObjectsOfClass.post(
            class_name=OnboardDiscoveryAttributes.__config_class__,
            object_dn=self._discovery_dn
        )

        return jobs.objects

    def wait_for_job_to_finish(self, job: 'Union[config.Object, str]', check_interval: int = 5, timeout: int = 300):
        """
        Waits for the  *Status* attribute to have a value other than *Pending Execution* and *Running*
        on the discovery job. An error is raised if the timeout is exceeded.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            check_interval: Poll interval in seconds to validate that the job finished.
            timeout: Timeout in seconds to wait for the job to finish.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=check_interval):
                if not self.is_in_progress(job=job):
                    return

        status = self._api.websdk.Config.Read.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status
        ).values[0]
        raise TimeoutError(
            f'Expected Network Discovery Job "{job_dn}" to finish within {timeout} seconds, but it is still '
            f'running. It has a status of "{status}."'
        )

@feature('Adaptable Onboard Discovery')
class AdaptableOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        devices_to_scan: list[Union[str, config.Object]] = None,
        folder_with_devices_to_scan: Union[str, config.Object] = None,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates an Adaptable Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            devices_to_scan: A list of :ref:`config_object` or :ref:`dn` of the device objects to scan.
            folder_with_devices_to_scan: A :ref:`config_object` or :ref:`dn` of the folder having device objects to scan.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}
        if devices_to_scan:
            devices_to_scan = [self._get_dn(d) for d in devices_to_scan]
            attributes[OnboardDiscoveryAttributes.device] = devices_to_scan
        if folder_with_devices_to_scan:
            attributes[OnboardDiscoveryAttributes.devices_folder] = self._get_dn(folder_with_devices_to_scan)
        return self._create(
            name=name,
            installation_type='Adaptable',
            driver_name='appadaptable',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )

@feature('AWS Onboard Discovery')
class AmazonOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        amazon_credential: Union[str, config.Object],
        amazon_account_ids: list[str] = None,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates an AWS Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            amazon_credential: A :ref:`config_object` or :ref:`dn` of the Amazon credential.
            amazon_account_ids: A list of Amazon account IDs.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}
        attributes.update(
            {
                OnboardDiscoveryAttributes.credential        : self._get_dn(amazon_credential),
                OnboardDiscoveryAttributes.amazon_account_ids: amazon_account_ids
            }
        )
        return self._create(
            name=name,
            installation_type='Amazon Web Services',
            driver_name='appamazon',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )

@feature('Azure Onboard Discovery')
class AzureOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        azure_credential: Union[str, config.Object],
        azure_application_id: str,
        azure_tenant_id: str,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates an Azure Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            azure_credential: A :ref:`config_object` or :ref:`dn` of the Azure credential.
            azure_application_id: Azure application ID.
            azure_tenant_id: Azure tenant ID.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}
        attributes.update(
            {
                OnboardDiscoveryAttributes.credential          : self._get_dn(azure_credential),
                OnboardDiscoveryAttributes.azure_application_id: azure_application_id,
                OnboardDiscoveryAttributes.azure_tenant_id     : azure_tenant_id,
            }
        )
        return self._create(
            name=name,
            installation_type='Azure Key Vault',
            driver_name='appazurekeyvault',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )

@feature('CAPI Onboard Discovery')
class CapiOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        devices_to_scan: list[Union[str, config.Object]] = None,
        extract_private_key_if_possible: bool = False,
        folder_with_devices_to_scan: Union[str, config.Object] = None,
        winrm_port: int = 5986,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates a CAPI (IIS Bindings) Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            extract_private_key_if_possible: If ``True``, extract the private key.
            winrm_port: WinRM port.
            devices_to_scan: A list of :ref:`config_object` or :ref:`dn` of the device objects to scan.
            folder_with_devices_to_scan: A :ref:`config_object` or :ref:`dn` of the folder having device objects to scan.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}
        attributes[OnboardDiscoveryAttributes.port] = winrm_port
        if devices_to_scan:
            devices_to_scan = [self._get_dn(d) for d in devices_to_scan]
            attributes[OnboardDiscoveryAttributes.device] = devices_to_scan
        if folder_with_devices_to_scan:
            attributes[OnboardDiscoveryAttributes.devices_folder] = self._get_dn(folder_with_devices_to_scan)
        if extract_private_key_if_possible:
            attributes[OnboardDiscoveryAttributes.extract_private_key] = "1"
        return self._create(
            name=name,
            installation_type='CAPI (IIS Bindings)',
            driver_name='appcapi',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )

@feature('Citrix NetScaler Onboard Discovery')
class NetScalerOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        certificates_to_import: Union[Literal['virtual', 'gateway', 'services', 'all'], int],
        keystore_and_private_key_credential: Union[str, config.Object] = None,
        devices_to_scan: list[Union[str, config.Object]] = None,
        extract_private_key_if_possible: bool = False,
        folder_with_devices_to_scan: Union[str, config.Object] = None,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates a Citrix NetScaler Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            devices_to_scan: A list of :ref:`config_object` or :ref:`dn` of the device objects to scan.
            folder_with_devices_to_scan: A :ref:`config_object` or :ref:`dn` of the folder having device objects to scan.
            extract_private_key_if_possible: If ``True``, extract the private key.
            certificates_to_import: Certificate profile from which to import certificates.
            keystore_and_private_key_credential: A :ref:`config_object` or :ref:`dn` of the keystore and private key credential.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}

        if isinstance(certificates_to_import, str):
            certificates_to_import = {
                'virtual' : 0,  # Only Virtual Servers
                'services': 1,  # Only Services and Groups
                'all'     : 2,  # All
                'gateway' : 3,  # Only Gateway Virtual Servers
            }.get(certificates_to_import)

        attributes.update(
            {
                OnboardDiscoveryAttributes.profiles_to_import: certificates_to_import
            }
        )
        if devices_to_scan:
            devices_to_scan = [self._get_dn(d) for d in devices_to_scan]
            attributes[OnboardDiscoveryAttributes.device] = devices_to_scan
        if folder_with_devices_to_scan:
            attributes[OnboardDiscoveryAttributes.devices_folder] = self._get_dn(folder_with_devices_to_scan)
        if extract_private_key_if_possible:
            attributes[OnboardDiscoveryAttributes.extract_private_key] = "1"
        if keystore_and_private_key_credential:
            attributes[OnboardDiscoveryAttributes.credential] = keystore_and_private_key_credential
        return self._create(
            name=name,
            installation_type='Citrix NetScaler',
            driver_name='appnetscaler',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )

@feature('F5 LTM Advanced Onboard Discovery')
class F5LtmAdvancedOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        certificates_to_import: Union[Literal['server', 'client', 'both'], int],
        port: int = 443,
        devices_to_scan: list[Union[str, config.Object]] = None,
        extract_private_key_if_possible: bool = False,
        folder_with_devices_to_scan: Union[str, config.Object] = None,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates an F5 LTM Advanced Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            devices_to_scan: A list of :ref:`config_object` or :ref:`dn` of the device objects to scan.
            folder_with_devices_to_scan: A :ref:`config_object` or :ref:`dn` of the folder having device objects to scan.
            extract_private_key_if_possible: If ``True``, extract the private key.
            certificates_to_import: Certificate profile from which to import certificates.
            port: Connection port.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}

        if isinstance(certificates_to_import, str):
            certificates_to_import = {
                'both'  : 0,
                'client': 1,
                'server': 2,
            }.get(certificates_to_import)

        attributes.update(
            {
                OnboardDiscoveryAttributes.profiles_to_import: certificates_to_import,
                OnboardDiscoveryAttributes.port              : port
            }
        )
        if devices_to_scan:
            devices_to_scan = [self._get_dn(d) for d in devices_to_scan]
            attributes[OnboardDiscoveryAttributes.device] = devices_to_scan
        if folder_with_devices_to_scan:
            attributes[OnboardDiscoveryAttributes.devices_folder] = self._get_dn(folder_with_devices_to_scan)
        if extract_private_key_if_possible:
            attributes[OnboardDiscoveryAttributes.extract_private_key] = "1"
        return self._create(
            name=name,
            installation_type='F5 LTM Advanced',
            driver_name='appf5ltmadvanced',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )

@feature('IBM DataPower Onboard Discovery')
class IbmDataPowerOnboardDiscovery(OnboardDiscoveryBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(
        self,
        name: str,
        certificates_to_import: Union[Literal['server', 'client', 'proxy', 'all'], int],
        port: int = 443,
        devices_to_scan: list[Union[str, config.Object]] = None,
        extract_private_key_if_possible: bool = False,
        folder_with_devices_to_scan: Union[str, config.Object] = None,
        attributes: dict = None,
        blackout: dict[str, List] = None,
        certificate_placement_folder: Union[str, config.Object] = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        days_of_week: list[str] = None,
        days_of_month: list[str] = None,
        days_of_year: list[str] = None,
        description: str = None,
        hour: int = None,
        priority: int = None,
        reschedule: bool = True,
        utc: str = '1',
        get_if_already_exists: bool = True
    ):
        """
        Creates an IBM DataPower Onboard Discovery job.

        Args:
            name: Name of the discovery job.
            devices_to_scan: A list of :ref:`config_object` or :ref:`dn` of the device objects to scan.
            folder_with_devices_to_scan: A :ref:`config_object` or :ref:`dn` of the folder having device objects to scan.
            extract_private_key_if_possible: If ``True``, extract the private key.
            certificates_to_import: Certificate profile from which to import certificates.
            port: Connection port.
            certificate_placement_folder: A :ref:`config_object` or :ref:`dn` of the folder to place the certificates.
            attributes: Additional attributes.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        attributes = attributes or {}

        if isinstance(certificates_to_import, str):
            certificates_to_import = {
                'all'   : 0,
                'client': 1,
                'proxy' : 2,
                'server': 3,
            }.get(certificates_to_import)

        attributes.update(
            {
                OnboardDiscoveryAttributes.profiles_to_import: certificates_to_import,
                OnboardDiscoveryAttributes.port              : port
            }
        )
        if devices_to_scan:
            devices_to_scan = [self._get_dn(d) for d in devices_to_scan]
            attributes[OnboardDiscoveryAttributes.device] = devices_to_scan
        if folder_with_devices_to_scan:
            attributes[OnboardDiscoveryAttributes.devices_folder] = self._get_dn(folder_with_devices_to_scan)
        if extract_private_key_if_possible:
            attributes[OnboardDiscoveryAttributes.extract_private_key] = "1"
        return self._create(
            name=name,
            installation_type='IBM DataPower',
            driver_name='appdatapower',
            attributes=attributes,
            blackout=blackout,
            certificate_placement_folder=certificate_placement_folder,
            contacts=contacts,
            days_of_week=days_of_week,
            days_of_month=days_of_month,
            days_of_year=days_of_year,
            description=description,
            hour=hour,
            priority=priority,
            reschedule=reschedule,
            utc=utc,
            get_if_already_exists=get_if_already_exists,
        )



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.tree_root import TreeRootAttributes

class BranchBaseAttributes(TreeRootAttributes, metaclass=IterableMeta):
    __config_class__ = "Branch Base"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.connection_base import ConnectionBaseAttributes
from pyvenafi.tpp.attributes.driver_base import DriverBaseAttributes

class CertificateTrustStoreBaseAttributes(ConnectionBaseAttributes, DriverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Certificate Trust Store Base"
    approver = Attribute('Approver')
    automatic_provisioning_disabled = Attribute('Automatic Provisioning Disabled')
    bundle = Attribute('Bundle')
    bundle_installed = Attribute('Bundle Installed')
    in_error = Attribute('In Error')
    interval = Attribute('Interval')
    stage = Attribute('Stage')
    status = Attribute('Status')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import (
    config,
    identity as ident,
)

class _Config(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Config')
        self.AddDnValue = self._AddDnValue(api_obj=api_obj, url=f'{self._url}/AddDnValue')
        self.AddPolicyValue = self._AddPolicyValue(api_obj=api_obj, url=f'{self._url}/AddPolicyValue')
        self.AddValue = self._AddValue(api_obj=api_obj, url=f'{self._url}/AddValue')
        self.ClearAttribute = self._ClearAttribute(api_obj=api_obj, url=f'{self._url}/ClearAttribute')
        self.ClearPolicyAttribute = self._ClearPolicyAttribute(api_obj=api_obj, url=f'{self._url}/ClearPolicyAttribute')
        self.ContainableClasses = self._ContainableClasses(api_obj=api_obj, url=f'{self._url}/ContainableClasses')
        self.CountObjects = self._CountObjects(api_obj=api_obj, url=f'{self._url}/CountObjects')
        self.Create = self._Create(api_obj=api_obj, url=f'{self._url}/Create')
        self.DefaultDN = self._DefaultDN(api_obj=api_obj, url=f'{self._url}/DefaultDN')
        self.Delete = self._Delete(api_obj=api_obj, url=f'{self._url}/Delete')
        self.DnToGuid = self._DnToGuid(api_obj=api_obj, url=f'{self._url}/DnToGuid')
        self.Enumerate = self._Enumerate(api_obj=api_obj, url=f'{self._url}/Enumerate')
        self.EnumerateAll = self._EnumerateAll(api_obj=api_obj, url=f'{self._url}/EnumerateAll')
        self.EnumerateObjectsDerivedFrom = self._EnumerateObjectsDerivedFrom(
            api_obj=api_obj,
            url=f'{self._url}/EnumerateObjectsDerivedFrom'
        )
        self.EnumeratePolicies = self._EnumeratePolicies(api_obj=api_obj, url=f'{self._url}/EnumeratePolicies')
        self.Find = self._Find(api_obj=api_obj, url=f'{self._url}/Find')
        self.FindContainers = self._FindContainers(api_obj=api_obj, url=f'{self._url}/FindContainers')
        self.FindObjectsOfClass = self._FindObjectsOfClass(api_obj=api_obj, url=f'{self._url}/FindObjectsOfClass')
        self.FindPolicy = self._FindPolicy(api_obj=api_obj, url=f'{self._url}/FindPolicy')
        self.GetHighestRevision = self._GetHighestRevision(api_obj=api_obj, url=f'{self._url}/GetHighestRevision')
        self.GetObjectTrustees = self._GetObjectTrustees(api_obj=api_obj, url=f'{self._url}/GetObjectTrustees')
        self.GetRevision = self._GetRevision(api_obj=api_obj, url=f'{self._url}/GetRevision')
        self.GuidToDn = self._GuidToDn(api_obj=api_obj, url=f'{self._url}/GuidToDn')
        self.IdInfo = self._IdInfo(api_obj=api_obj, url=f'{self._url}/IdInfo')
        self.IsValid = self._IsValid(api_obj=api_obj, url=f'{self._url}/IsValid')
        self.MutateObject = self._MutateObject(api_obj=api_obj, url=f'{self._url}/MutateObject')
        self.Read = self._Read(api_obj=api_obj, url=f'{self._url}/Read')
        self.ReadAll = self._ReadAll(api_obj=api_obj, url=f'{self._url}/ReadAll')
        self.ReadDn = self._ReadDn(api_obj=api_obj, url=f'{self._url}/ReadDn')
        self.ReadDnReferences = self._ReadDnReferences(api_obj=api_obj, url=f'{self._url}/ReadDnReferences')
        self.ReadEffectivePolicy = self._ReadEffectivePolicy(api_obj=api_obj, url=f'{self._url}/ReadEffectivePolicy')
        self.ReadPolicy = self._ReadPolicy(api_obj=api_obj, url=f'{self._url}/ReadPolicy')
        self.RemoveDnValue = self._RemoveDnValue(api_obj=api_obj, url=f'{self._url}/RemoveDnValue')
        self.RemovePolicyValue = self._RemovePolicyValue(api_obj=api_obj, url=f'{self._url}/RemovePolicyValue')
        self.RemoveValue = self._RemoveValue(api_obj=api_obj, url=f'{self._url}/GetObjectTrustees')
        self.RenameObject = self._RenameObject(api_obj=api_obj, url=f'{self._url}/RenameObject')
        self.Write = self._Write(api_obj=api_obj, url=f'{self._url}/Write')
        self.WriteDn = self._WriteDn(api_obj=api_obj, url=f'{self._url}/WriteDn')
        self.WritePolicy = self._WritePolicy(api_obj=api_obj, url=f'{self._url}/WritePolicy')

    class _AddDnValue(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, value: str):
            body = {
                'ObjectDN'     : object_dn,
                'AttributeName': attribute_name,
                'Value'        : value
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _AddPolicyValue(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, class_name: str, value: str, locked: bool):
            body = {
                'ObjectDN'     : object_dn,
                'AttributeName': attribute_name,
                'Class'        : class_name,
                'Value'        : value,
                'Locked'       : locked
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _AddValue(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, value: str):
            body = {
                'ObjectDN'     : object_dn,
                'AttributeName': attribute_name,
                'Value'        : value,
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ClearAttribute(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str):
            body = {
                'ObjectDN'     : object_dn,
                'AttributeName': attribute_name
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ClearPolicyAttribute(WebSdkEndpoint):
        def post(self, object_dn: str, class_name: str, attribute_name: str):
            body = {
                'ObjectDN'     : object_dn,
                'Class'        : class_name,
                'AttributeName': attribute_name
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ContainableClasses(WebSdkEndpoint):
        def post(self, object_dn: str):
            body = {
                'ObjectDN': object_dn
            }

            class Output(WebSdkOutputModel):
                class_names: list[str] = ApiField(alias='ClassNames', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CountObjects(WebSdkEndpoint):
        def post(self, object_dn: str, type_name: str, recursive: bool = False, pattern: str = None):
            body = {
                'ObjectDN' : object_dn,
                'Type'     : type_name,
                'Pattern'  : pattern,
                'Recursive': recursive
            }

            class Output(WebSdkOutputModel):
                count: int = ApiField(alias='Count')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _Create(WebSdkEndpoint):
        def post(self, object_dn: str, class_name: str, name_attribute_list: list[config.NameAttribute]):
            body = {
                "ObjectDN"         : object_dn,
                "Class"            : class_name,
                "NameAttributeList": name_attribute_list
            }

            class Output(WebSdkOutputModel):
                object: config.Object = ApiField(alias='Object')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DefaultDN(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                default_dn: str = ApiField(alias='DefaultDN')
                result: int = ApiField(alias='Result')

            return generate_output(response=self._get(), output_cls=Output)

    class _Delete(WebSdkEndpoint):
        def post(self, object_dn: str, recursive: bool = False):
            body = {
                "ObjectDN" : object_dn,
                "Recursive": recursive
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DnToGuid(WebSdkEndpoint):
        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn,
            }

            class Output(WebSdkOutputModel):
                class_name: str = ApiField(alias='ClassName')
                guid: str = ApiField(alias='GUID')
                revision: str = ApiField(alias='Revision')
                hierarchical_guid: str = ApiField(alias='HierarchicalGUID')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _Enumerate(WebSdkEndpoint):
        def post(self, object_dn: str = None, recursive: bool = False, pattern: str = None):
            body = {
                "ObjectDN" : object_dn,
                "Recursive": recursive,
                "Pattern"  : pattern
            }

            class Output(WebSdkOutputModel):
                objects: list[config.Object] = ApiField(alias='Objects', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateAll(WebSdkEndpoint):
        def post(self, pattern: str):
            body = {
                "Pattern": pattern
            }

            class Output(WebSdkOutputModel):
                objects: list[config.Object] = ApiField(alias='Objects', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateObjectsDerivedFrom(WebSdkEndpoint):
        def post(self, derived_from: str, pattern: str = None):
            body = {
                "DerivedFrom": derived_from,
                "Pattern"    : pattern
            }

            class Output(WebSdkOutputModel):
                objects: list[config.Object] = ApiField(alias='Objects', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumeratePolicies(WebSdkEndpoint):
        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn
            }

            class Output(WebSdkOutputModel):
                policies: list[config.Policy] = ApiField(alias='Policies', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _Find(WebSdkEndpoint):
        def post(self, pattern: str, attribute_names: list[str] = None):
            body = {
                "Pattern"       : pattern,
                "AttributeNames": attribute_names
            }

            class Output(WebSdkOutputModel):
                objects: list[config.Object] = ApiField(alias='Objects', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _FindContainers(WebSdkEndpoint):
        def post(self, object_dn: str, recursive: bool = False):
            body = {
                "ObjectDN" : object_dn,
                "Recursive": recursive
            }

            class Output(WebSdkOutputModel):
                objects: list[config.Object] = ApiField(alias='Objects', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _FindObjectsOfClass(WebSdkEndpoint):
        def post(
            self, classes: str = None, class_name: str = None, object_dn: str = None, pattern: str = None,
            recursive: bool = False
        ):
            if not (classes or class_name):
                raise AssertionError('One of "classes" or "class_name" parameters must be provided.')
            body = {
                "Classes"  : classes,
                "Class"    : class_name,
                'ObjectDN' : object_dn,
                'Pattern'  : pattern,
                'Recursive': recursive
            }

            class Output(WebSdkOutputModel):
                objects: list[config.Object] = ApiField(alias='Objects', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _FindPolicy(WebSdkEndpoint):
        def post(self, object_dn: str, class_name: str, attribute_name: str):
            body = {
                "ObjectDN"     : object_dn,
                "Class"        : class_name,
                "AttributeName": attribute_name
            }

            class Output(WebSdkOutputModel):
                locked: bool = ApiField(alias='Locked')
                policy_dn: str = ApiField(alias='PolicyDN')
                values: list[str] = ApiField(alias='Values', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetHighestRevision(WebSdkEndpoint):
        def post(self, object_dn: str, classes: str = None):
            body = {
                "ObjectDN": object_dn,
                'Classes' : classes
            }

            class Output(WebSdkOutputModel):
                revision: int = ApiField(alias='Revision')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetRevision(WebSdkEndpoint):
        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn
            }

            class Output(WebSdkOutputModel):
                revision: int = ApiField(alias='Revision')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetObjectTrustees(WebSdkEndpoint):
        def post(
            self,
            recursive: bool = None,
            include_derived_classes: bool = None,
            locked: bool = None,
            pattern: str = None,
            object_pattern: str = None,
            derived_from: str = None,
            object_id: int = None,
            object_dn: str = None,
            type: str = None,
            new_object_dn: str = None,
            object_guid: str = None,
            classes: list[str] = None,
            className: str = None,
            attribute_name: str = None,
            attribute_names: list[str] = None,
            value: str = None,
            values: list[str] = None,
            attribute_data: dict[str, any] = None,
            reference_attribute_name: str = None,
            name_attribute_list: dict[str, any] = None,
        ):
            body = {
                "Recursive"             : recursive,
                "IncludeDerivedClasses" : include_derived_classes,
                "Locked"                : locked,
                "Pattern"               : pattern,
                "ObjectPattern"         : object_pattern,
                "DerivedFrom"           : derived_from,
                "ObjectID"              : object_id,
                "ObjectDN"              : object_dn,
                "Type"                  : type,
                "NewObjectDN"           : new_object_dn,
                "ObjectGUID"            : object_guid,
                "Classes"               : classes,
                "Class"                 : className,
                "AttributeName"         : attribute_name,
                "AttributeNames"        : attribute_names,
                "Value"                 : value,
                "Values"                : values,
                "AttributeData"         : attribute_data,
                "ReferenceAttributeName": reference_attribute_name,
                "NameAttributeList"     : name_attribute_list,
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))
                count: int = ApiField(alias='Count')
                object: config.Object = ApiField(alias='Object')
                objects: list[config.Object] = ApiField(alias='Objects')
                name_values: list[config.NameValues] = ApiField(alias='NameValues')
                class_definition: config.ClassDefinition = ApiField(alias='ClassDefinition')
                class_name: str = ApiField(alias='ClassName')
                class_names: list[str] = ApiField(alias='ClassNames')
                object_dn: str = ApiField(alias='ObjectDN')
                guid: str = ApiField(alias='GUID')
                hierarchical_guid: str = ApiField(alias='HierarchicalGUID')
                class_definitions: list[config.ClassDefinition] = ApiField(alias='ClassDefinitions')
                attribute_definition: config.AttributeDefinition = ApiField(alias='AttributeDefinition')
                attribute_definitions: list[config.AttributeDefinition] = ApiField(alias='AttributeDefinitions')
                values: list[str] = ApiField(alias='Values')
                rights: int = ApiField(alias='Rights')
                policies: list[config.Policy] = ApiField(alias='Policies')
                policy_dn: str = ApiField(alias='PolicyDN')
                locked: bool = ApiField(alias='Locked')
                overridden: bool = ApiField(alias='Overridden')
                revision: int = ApiField(alias='Revision')
                identities: list[ident.Identity] = ApiField(alias='Identities')
                default_dn: str = ApiField(alias='DefaultDN')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _GuidToDn(WebSdkEndpoint):
        def post(self, object_guid: str):
            body = {
                "ObjectGUID": object_guid
            }

            class Output(WebSdkOutputModel):
                object_dn: str = ApiField(alias='ObjectDN')
                class_name: str = ApiField(alias='ClassName')
                revision: str = ApiField(alias='Revision')
                hierarchical_guid: str = ApiField(alias='HierarchicalGUID')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _IdInfo(WebSdkEndpoint):
        def post(self, object_id: str):
            body = {
                "ObjectID": object_id
            }

            class Output(WebSdkOutputModel):
                guid: str = ApiField(alias='GUID')
                class_name: str = ApiField(alias='ClassName')
                revision: str = ApiField(alias='Revision')
                hierarchical_guid: str = ApiField(alias='HierarchicalGUID')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _IsValid(WebSdkEndpoint):
        def post(self, object_dn: str = None, object_guid: str = None):
            body = {
                "ObjectGUID": object_guid,
                "ObjectDN"  : object_dn
            }

            class Output(WebSdkOutputModel):
                object: config.Object = ApiField(alias='Object')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _MutateObject(WebSdkEndpoint):
        def post(self, object_dn: str, class_name: str):
            body = {
                "ObjectDN": object_dn,
                "Class"   : class_name
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _Read(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name
            }

            class Output(WebSdkOutputModel):
                values: list[str] = ApiField(alias='Values', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ReadAll(WebSdkEndpoint):
        def post(self, object_dn: str):
            body = {
                "ObjectDN": object_dn
            }

            class Output(WebSdkOutputModel):
                name_values: list[config.NameValues[str]] = ApiField(alias='NameValues', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ReadDn(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name
            }

            class Output(WebSdkOutputModel):
                values: list[str] = ApiField(alias='Values', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ReadDnReferences(WebSdkEndpoint):
        def post(self, object_dn: str, reference_attribute_name: str, attribute_name: str):
            body = {
                "ObjectDN"              : object_dn,
                "ReferenceAttributeName": reference_attribute_name,
                "AttributeName"         : attribute_name
            }

            class Output(WebSdkOutputModel):
                values: list[str] = ApiField(alias='Values', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ReadEffectivePolicy(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name
            }

            class Output(WebSdkOutputModel):
                values: list[str] = ApiField(alias='Values', default_factory=list)
                locked: bool = ApiField(alias='Locked')
                overridden: bool = ApiField(alias='Overridden')
                policy_dn: str = ApiField(alias='PolicyDN')
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ReadPolicy(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, class_name: str):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name,
                "Class"        : class_name
            }

            class Output(WebSdkOutputModel):
                locked: bool = ApiField(alias='Locked')
                values: list[str] = ApiField(alias='Values', default_factory=list)
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RemoveDnValue(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, value: str):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name,
                "Value"        : value
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RemovePolicyValue(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, class_name: str, value: str):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name,
                "Class"        : class_name,
                "Value"        : value
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RemoveValue(WebSdkEndpoint):
        def post(
            self,
            recursive: bool = None,
            include_derived_classes: bool = None,
            locked: bool = None,
            pattern: str = None,
            object_pattern: str = None,
            derived_from: str = None,
            object_id: int = None,
            object_dn: str = None,
            type: str = None,
            new_object_dn: str = None,
            object_guid: str = None,
            classes: list[str] = None,
            className: str = None,
            attribute_name: str = None,
            attribute_names: list[str] = None,
            value: str = None,
            values: list[str] = None,
            attribute_data: dict[str, any] = None,
            reference_attribute_name: str = None,
            name_attribute_list: dict[str, any] = None,
        ):
            body = {
                "Recursive"             : recursive,
                "IncludeDerivedClasses" : include_derived_classes,
                "Locked"                : locked,
                "Pattern"               : pattern,
                "ObjectPattern"         : object_pattern,
                "DerivedFrom"           : derived_from,
                "ObjectID"              : object_id,
                "ObjectDN"              : object_dn,
                "Type"                  : type,
                "NewObjectDN"           : new_object_dn,
                "ObjectGUID"            : object_guid,
                "Classes"               : classes,
                "Class"                 : className,
                "AttributeName"         : attribute_name,
                "AttributeNames"        : attribute_names,
                "Value"                 : value,
                "Values"                : values,
                "AttributeData"         : attribute_data,
                "ReferenceAttributeName": reference_attribute_name,
                "NameAttributeList"     : name_attribute_list,
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))
                count: int = ApiField(alias='Count')
                object: config.Object = ApiField(alias='Object')
                objects: list[config.Object] = ApiField(alias='Objects')
                name_values: list[config.NameValues] = ApiField(alias='NameValues')
                class_definition: config.ClassDefinition = ApiField(alias='ClassDefinition')
                class_name: str = ApiField(alias='ClassName')
                class_names: list[str] = ApiField(alias='ClassNames')
                object_dn: str = ApiField(alias='ObjectDN')
                guid: str = ApiField(alias='GUID')
                hierarchical_guid: str = ApiField(alias='HierarchicalGUID')
                class_definitions: list[config.ClassDefinition] = ApiField(alias='ClassDefinitions')
                attribute_definition: config.AttributeDefinition = ApiField(alias='AttributeDefinition')
                attribute_definitions: list[config.AttributeDefinition] = ApiField(alias='AttributeDefinitions')
                values: list[str] = ApiField(alias='Values')
                rights: int = ApiField(alias='Rights')
                policies: list[config.Policy] = ApiField(alias='Policies')
                policy_dn: str = ApiField(alias='PolicyDN')
                locked: bool = ApiField(alias='Locked')
                overridden: bool = ApiField(alias='Overridden')
                revision: int = ApiField(alias='Revision')
                identities: list[ident.Identity] = ApiField(alias='Identities')
                default_dn: str = ApiField(alias='DefaultDN')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _RenameObject(WebSdkEndpoint):
        def post(self, object_dn: str, new_object_dn: str):
            body = {
                "ObjectDN"   : object_dn,
                "NewObjectDN": new_object_dn
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _Write(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_data: list[dict[str, list[str]]]):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeData": attribute_data
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _WriteDn(WebSdkEndpoint):
        def post(self, object_dn: str, attribute_name: str, values: list[str]):
            body = {
                "ObjectDN"     : object_dn,
                "AttributeName": attribute_name,
                "Values"       : values
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _WritePolicy(WebSdkEndpoint):
        def post(
            self,
            object_dn: str,
            class_name: str,
            attribute_name: str,
            locked: bool = False,
            values: list[str] = None
        ):
            body = {
                "ObjectDN"     : object_dn,
                "Class"        : class_name,
                "AttributeName": attribute_name,
                "Locked"       : locked,
                "Values"       : values
            }

            class Output(WebSdkOutputModel):
                result: config.Result = ApiField(alias='Result', converter=lambda x: config.Result(code=x))

            return generate_output(response=self._post(data=body), output_cls=Output)



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    CloudApiEndpoint,
    CloudApiOutputModel,
    generate_output,
)
from pyvenafi.cloud.api.models import outagedetection_service
from typing import (
    Any,
    Literal,
)

class _outagedetection_service:
    def __init__(self, api_obj):
        self.outagedetection = self._outagedetection(api_obj=api_obj)

    class _outagedetection(CloudApiEndpoint):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='outagedetection')
            self.v1 = self._v1(api_obj=self._api_obj, url=f'{self._url}/v1')

        class _v1(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.applications = self._applications(api_obj=self._api_obj, url=f'{self._url}/applications')
                self.certificateaggregates = self._certificateaggregates(
                    api_obj=self._api_obj,
                    url=f'{self._url}/certificateaggregates'
                )
                self.certificateinstances = self._certificateinstances(
                    api_obj=self._api_obj,
                    url=f'{self._url}/certificateinstances'
                )
                self.certificateinstancesearch = self._certificateinstancesearch(
                    api_obj=self._api_obj, url=f'{self._url}/certificateinstancesearch'
                )
                self.certificaterequests = self._certificaterequests(
                    api_obj=self._api_obj,
                    url=f'{self._url}/certificaterequests'
                )
                self.certificaterequestssearch = self._certificaterequestssearch(
                    api_obj=self._api_obj, url=f'{self._url}/certificaterequestssearch'
                )
                self.certificates = self._certificates(api_obj=self._api_obj, url=f'{self._url}/certificates')
                self.certificatesearch = self._certificatesearch(
                    api_obj=self._api_obj,
                    url=f'{self._url}/certificatesearch'
                )
                self.inventorymonitoringconfig = self._inventorymonitoringconfig(
                    api_obj=self._api_obj, url=f'{self._url}/inventorymonitoringconfig'
                )
                self.savedsearches = self._savedsearches(api_obj=self._api_obj, url=f'{self._url}/savedsearches')

            class _applications(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.certificates = self._certificates(api_obj=self._api_obj, url=f'{self._url}/certificates')
                    self.name = self._name(api_obj=self._api_obj, url=f'{self._url}/name')

                def ID(self, id: str):
                    return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                def NAME(self, name: str):
                    return self._NAME(api_obj=self._api_obj, url=f'{self._url}/{name}')

                def get(
                    self,
                    issuingTemplateAssigned: bool,
                    ownerDetails: bool,
                    ownershipCheck: bool,
                    ownershipTree: bool
                ):
                    data = {
                        'issuingTemplateAssigned': issuingTemplateAssigned,
                        'ownerDetails'           : ownerDetails,
                        'ownershipCheck'         : ownershipCheck,
                        'ownershipTree'          : ownershipTree,
                    }

                    class Output(CloudApiOutputModel):
                        ApplicationResponse: outagedetection_service.ApplicationResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'ApplicationResponse'
                        }
                    )

                def post(self, ApplicationRequest: outagedetection_service.ApplicationRequest):
                    data = {**ApplicationRequest.dict()}

                    class Output(CloudApiOutputModel):
                        ApplicationResponse: outagedetection_service.ApplicationResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'ApplicationResponse'
                        }
                    )

                class _ID(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                        self.invitations = self._invitations(api_obj=self._api_obj, url=f'{self._url}/invitations')
                        self.scanaficonfiguration = self._scanaficonfiguration(
                            api_obj=self._api_obj,
                            url=f'{self._url}/scanaficonfiguration'
                        )

                    def delete(self):
                        class Output(CloudApiOutputModel):
                            pass

                        return generate_output(output_cls=Output, response=self._delete(params={}))

                    def get(self, ownershipTree: bool):
                        data = {
                            'ownershipTree': ownershipTree,
                        }

                        class Output(CloudApiOutputModel):
                            ApplicationInformation: outagedetection_service.ApplicationInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params=data),
                            rc_mapping={
                                200: 'ApplicationInformation'
                            }
                        )

                    def put(self, ApplicationRequest: outagedetection_service.ApplicationRequest):
                        data = {**ApplicationRequest.dict()}

                        class Output(CloudApiOutputModel):
                            ApplicationInformation: outagedetection_service.ApplicationInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'ApplicationInformation'
                            }
                        )

                    class _invitations(CloudApiEndpoint):
                        def post(self, InvitationRequest: outagedetection_service.InvitationRequest):
                            data = {**InvitationRequest.dict()}

                            class Output(CloudApiOutputModel):
                                InvitationResponse: outagedetection_service.InvitationResponse

                            return generate_output(
                                output_cls=Output,
                                response=self._post(data=data),
                                rc_mapping={
                                    201: 'InvitationResponse'
                                }
                            )

                    class _scanaficonfiguration(CloudApiEndpoint):
                        def get(self):
                            class Output(CloudApiOutputModel):
                                ScanafiConfigResponseV1: outagedetection_service.ScanafiConfigResponseV1

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params={}),
                                rc_mapping={
                                    200: 'ScanafiConfigResponseV1'
                                }
                            )

                class _NAME(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                        self.certificateissuingtemplates = self._certificateissuingtemplates(
                            api_obj=self._api_obj, url=f'{self._url}/certificateissuingtemplates'
                        )

                    class _certificateissuingtemplates(CloudApiEndpoint):
                        def __init__(self, *args, **kwargs):
                            super().__init__(*args, **kwargs)

                        def ALIAS(self, alias: str):
                            return self._ALIAS(api_obj=self._api_obj, url=f'{self._url}/{alias}')

                        class _ALIAS(CloudApiEndpoint):
                            def get(self):
                                class Output(CloudApiOutputModel):
                                    CertificateIssuingTemplateInformation: outagedetection_service.CertificateIssuingTemplateInformation

                                return generate_output(
                                    output_cls=Output,
                                    response=self._get(params={}),
                                    rc_mapping={
                                        200: 'CertificateIssuingTemplateInformation'
                                    }
                                )

                class _certificates(CloudApiEndpoint):
                    def patch(self, ApplicationsAssignRequest: outagedetection_service.ApplicationsAssignRequest):
                        data = {**ApplicationsAssignRequest.dict()}

                        class Output(CloudApiOutputModel):
                            ApplicationsAssignResponse: outagedetection_service.ApplicationsAssignResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._patch(data=data),
                            rc_mapping={
                                200: 'ApplicationsAssignResponse'
                            }
                        )

                class _name(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)

                    def NAME(self, name: str):
                        return self._NAME(api_obj=self._api_obj, url=f'{self._url}/{name}')

                    class _NAME(CloudApiEndpoint):
                        def get(self, ownershipTree: bool):
                            data = {
                                'ownershipTree': ownershipTree,
                            }

                            class Output(CloudApiOutputModel):
                                ApplicationInformation: outagedetection_service.ApplicationInformation

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params=data),
                                rc_mapping={
                                    200: 'ApplicationInformation'
                                }
                            )

            class _certificateaggregates(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.quota = self._quota(api_obj=self._api_obj, url=f'{self._url}/quota')
                    self.range = self._range(api_obj=self._api_obj, url=f'{self._url}/range')
                    self.unassigned = self._unassigned(api_obj=self._api_obj, url=f'{self._url}/unassigned')

                def NAME(self, name: str):
                    return self._NAME(api_obj=self._api_obj, url=f'{self._url}/{name}')

                def post(self, CertificateAggregationsRequest: outagedetection_service.CertificateAggregationsRequest):
                    data = {**CertificateAggregationsRequest.dict()}

                    class Output(CloudApiOutputModel):
                        FacetResponse: outagedetection_service.FacetResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            200: 'FacetResponse'
                        }
                    )

                class _NAME(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            CertificateAggregatesResponse: outagedetection_service.CertificateAggregatesResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CertificateAggregatesResponse'
                            }
                        )

                class _quota(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            CertificateInstallationsUsageResponse: outagedetection_service.CertificateInstallationsUsageResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CertificateInstallationsUsageResponse'
                            }
                        )

                class _range(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            CertificateAggregatesRangeResponse: outagedetection_service.CertificateAggregatesRangeResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CertificateAggregatesRangeResponse'
                            }
                        )

                class _unassigned(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            UnassignedCertificateAggregatesResponse: outagedetection_service.UnassignedCertificateAggregatesResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'UnassignedCertificateAggregatesResponse'
                            }
                        )

            class _certificateinstances(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.validation = self._validation(api_obj=self._api_obj, url=f'{self._url}/validation')

                def ID(self, id: str):
                    return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                def get(
                    self,
                    hostname: str,
                    ipAddress: str,
                    limit: int,
                    source: Literal[
                        'AWS_DISCOVERY', 'AZURE_DISCOVERY', 'DOMAIN_SCAN', 'EXTERNAL_CA_IMPORT', 'EXTERNAL_SCAN', 'FILE_IMPORT', 'GCP_DISCOVERY', 'KUBERNETES_DISCOVERY', 'MACHINE_DISCOVERY', 'ON_PREM_CA_IMPORT', 'SMART_SCAN_EXTERNAL', 'SMART_SCAN_INTERNAL', 'SMART_VALIDATION_EXTERNAL', 'SMART_VALIDATION_INTERNAL', 'TRUSTNET_SCAN', 'UNKNOWN', 'USER_IMPORTED', 'USER_PROVIDED', 'USER_SCAN']
                ):
                    data = {
                        'hostname' : hostname,
                        'ipAddress': ipAddress,
                        'limit'    : limit,
                        'source'   : source,
                    }

                    class Output(CloudApiOutputModel):
                        ExtendedCertificateInstanceResponse: outagedetection_service.ExtendedCertificateInstanceResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'ExtendedCertificateInstanceResponse'
                        }
                    )

                class _ID(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            ExtendedCertificateInstanceInformation: outagedetection_service.ExtendedCertificateInstanceInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'ExtendedCertificateInstanceInformation'
                            }
                        )

                class _validation(CloudApiEndpoint):
                    def post(
                        self,
                        CertificateInstanceValidationRequest: outagedetection_service.CertificateInstanceValidationRequest
                    ):
                        data = {**CertificateInstanceValidationRequest.dict()}

                        class Output(CloudApiOutputModel):
                            Response: outagedetection_service.Response

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                202: 'Response'
                            }
                        )

            class _certificateinstancesearch(CloudApiEndpoint):
                def post(self, Filter: outagedetection_service.Filter):
                    data = {**Filter.dict()}

                    class Output(CloudApiOutputModel):
                        ExtendedCertificateInstanceResponse: outagedetection_service.ExtendedCertificateInstanceResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            200: 'ExtendedCertificateInstanceResponse'
                        }
                    )

            class _certificaterequests(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.validation = self._validation(api_obj=self._api_obj, url=f'{self._url}/validation')

                def ID(self, id: str):
                    return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                def get(self):
                    class Output(CloudApiOutputModel):
                        CertificateRequestResponse: outagedetection_service.CertificateRequestResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'CertificateRequestResponse'
                        }
                    )

                def post(self, CertificateRequestRequest: outagedetection_service.CertificateRequestRequest):
                    data = {**CertificateRequestRequest.dict()}

                    class Output(CloudApiOutputModel):
                        CertificateRequestResponse: outagedetection_service.CertificateRequestResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'CertificateRequestResponse'
                        }
                    )

                class _ID(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                        self.resubmission = self._resubmission(api_obj=self._api_obj, url=f'{self._url}/resubmission')

                    def get(self):
                        class Output(CloudApiOutputModel):
                            CertificateRequestInformation: outagedetection_service.CertificateRequestInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CertificateRequestInformation'
                            }
                        )

                    class _resubmission(CloudApiEndpoint):
                        def post(
                            self,
                            CertificateRequestResubmissionRequest: outagedetection_service.CertificateRequestResubmissionRequest
                        ):
                            data = {**CertificateRequestResubmissionRequest.dict()}

                            class Output(CloudApiOutputModel):
                                CertificateRequestResponse: outagedetection_service.CertificateRequestResponse

                            return generate_output(
                                output_cls=Output,
                                response=self._post(data=data),
                                rc_mapping={
                                    200: 'CertificateRequestResponse'
                                }
                            )

                class _validation(CloudApiEndpoint):
                    def post(self, CertificateRequestRequest: outagedetection_service.CertificateRequestRequest):
                        data = {**CertificateRequestRequest.dict()}

                        class Output(CloudApiOutputModel):
                            CertificationRequestInformation: outagedetection_service.CertificationRequestInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                201: 'CertificationRequestInformation'
                            }
                        )

            class _certificaterequestssearch(CloudApiEndpoint):
                def post(self, Filter: outagedetection_service.Filter):
                    data = {**Filter.dict()}

                    class Output(CloudApiOutputModel):
                        CertificateRequestDocumentResponse: outagedetection_service.CertificateRequestDocumentResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            200: 'CertificateRequestDocumentResponse'
                        }
                    )

            class _certificates(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.deletion = self._deletion(api_obj=self._api_obj, url=f'{self._url}/deletion')
                    self.managed = self._managed(api_obj=self._api_obj, url=f'{self._url}/managed')
                    self.recovery = self._recovery(api_obj=self._api_obj, url=f'{self._url}/recovery')
                    self.retirement = self._retirement(api_obj=self._api_obj, url=f'{self._url}/retirement')
                    self.validation = self._validation(api_obj=self._api_obj, url=f'{self._url}/validation')

                def ID(self, id: str):
                    return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                def get(self, excludeSupersededInstances: bool, limit: Any, ownershipTree: bool, subject: str):
                    data = {
                        'excludeSupersededInstances': excludeSupersededInstances,
                        'limit'                     : limit,
                        'ownershipTree'             : ownershipTree,
                        'subject'                   : subject,
                    }

                    class Output(CloudApiOutputModel):
                        CertificateResponse: outagedetection_service.CertificateResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'CertificateResponse'
                        }
                    )

                def post(self, CertificateImportRequest: outagedetection_service.CertificateImportRequest):
                    data = {**CertificateImportRequest.dict()}

                    class Output(CloudApiOutputModel):
                        CertificateImportResponse: outagedetection_service.CertificateImportResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'CertificateImportResponse'
                        }
                    )

                class _ID(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                        self.contents = self._contents(api_obj=self._api_obj, url=f'{self._url}/contents')
                        self.keystore = self._keystore(api_obj=self._api_obj, url=f'{self._url}/keystore')

                    def get(self, excludeSupersededInstances: bool, ownershipTree: bool):
                        data = {
                            'excludeSupersededInstances': excludeSupersededInstances,
                            'ownershipTree'             : ownershipTree,
                        }

                        class Output(CloudApiOutputModel):
                            ExtendedCertificateInformation: outagedetection_service.ExtendedCertificateInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params=data),
                            rc_mapping={
                                200: 'ExtendedCertificateInformation'
                            }
                        )

                    class _contents(CloudApiEndpoint):
                        def get(
                            self,
                            Accept: Literal['application/octet-stream', 'text/plain'],
                            chainOrder: Literal['EE_FIRST', 'EE_ONLY', 'ROOT_FIRST'],
                            format: Literal['DER', 'PEM']
                        ):
                            data = {
                                'Accept'    : Accept,
                                'chainOrder': chainOrder,
                                'format'    : format,
                            }

                            class Output(CloudApiOutputModel):
                                str: str

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params=data),
                                rc_mapping={
                                    200: 'str'
                                }
                            )

                    class _keystore(CloudApiEndpoint):
                        def post(self, CertificateKeystoreRequest: outagedetection_service.CertificateKeystoreRequest):
                            data = {**CertificateKeystoreRequest.dict()}

                            class Output(CloudApiOutputModel):
                                str: str

                            return generate_output(
                                output_cls=Output,
                                response=self._post(data=data),
                                rc_mapping={
                                    200: 'str'
                                }
                            )

                class _deletion(CloudApiEndpoint):
                    def post(self, CertificateDeletionRequest: outagedetection_service.CertificateDeletionRequest):
                        data = {**CertificateDeletionRequest.dict()}

                        class Output(CloudApiOutputModel):
                            pass

                        return generate_output(output_cls=Output, response=self._post(data=data))

                class _managed(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)

                    def ID(self, id: str):
                        return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                    class _ID(CloudApiEndpoint):
                        def get(self, excludeSupersededInstances: bool, ownershipTree: bool):
                            data = {
                                'excludeSupersededInstances': excludeSupersededInstances,
                                'ownershipTree'             : ownershipTree,
                            }

                            class Output(CloudApiOutputModel):
                                CertificateResponse: outagedetection_service.CertificateResponse

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params=data),
                                rc_mapping={
                                    200: 'CertificateResponse'
                                }
                            )

                class _recovery(CloudApiEndpoint):
                    def post(self, CertificateRecoveryRequest: outagedetection_service.CertificateRecoveryRequest):
                        data = {**CertificateRecoveryRequest.dict()}

                        class Output(CloudApiOutputModel):
                            CertificateResponse: outagedetection_service.CertificateResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                200: 'CertificateResponse'
                            }
                        )

                class _retirement(CloudApiEndpoint):
                    def post(self, CertificateRetirementRequest: outagedetection_service.CertificateRetirementRequest):
                        data = {**CertificateRetirementRequest.dict()}

                        class Output(CloudApiOutputModel):
                            CertificateResponse: outagedetection_service.CertificateResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                200: 'CertificateResponse'
                            }
                        )

                class _validation(CloudApiEndpoint):
                    def post(self, CertificateValidationRequest: outagedetection_service.CertificateValidationRequest):
                        data = {**CertificateValidationRequest.dict()}

                        class Output(CloudApiOutputModel):
                            CertificateValidationResponse: outagedetection_service.CertificateValidationResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                202: 'CertificateValidationResponse'
                            }
                        )

            class _certificatesearch(CloudApiEndpoint):
                def post(self, excludeSupersededInstances: bool, ownershipTree: bool):
                    data = {
                        'excludeSupersededInstances': excludeSupersededInstances,
                        'ownershipTree'             : ownershipTree,
                    }

                    class Output(CloudApiOutputModel):
                        CertificateResponse: outagedetection_service.CertificateResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            200: 'CertificateResponse'
                        }
                    )

            class _inventorymonitoringconfig(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                def TYPE(self, type: str):
                    return self._TYPE(api_obj=self._api_obj, url=f'{self._url}/{type}')

                class _TYPE(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                        self.scheduler = self._scheduler(api_obj=self._api_obj, url=f'{self._url}/scheduler')

                    def get(self):
                        class Output(CloudApiOutputModel):
                            InventoryMonitoringConfigurationResponse: outagedetection_service.InventoryMonitoringConfigurationResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'InventoryMonitoringConfigurationResponse'
                            }
                        )

                    def put(
                        self,
                        InventoryMonitoringConfigRequest: outagedetection_service.InventoryMonitoringConfigRequest
                    ):
                        data = {**InventoryMonitoringConfigRequest.dict()}

                        class Output(CloudApiOutputModel):
                            InventoryMonitoringConfigurationResponse: outagedetection_service.InventoryMonitoringConfigurationResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'InventoryMonitoringConfigurationResponse'
                            }
                        )

                    class _scheduler(CloudApiEndpoint):
                        def put(self, runNow: bool):
                            data = {
                                'runNow': runNow,
                            }

                            class Output(CloudApiOutputModel):
                                InventoryMonitoringSchedulerInformation: outagedetection_service.InventoryMonitoringSchedulerInformation

                            return generate_output(
                                output_cls=Output,
                                response=self._put(data=data),
                                rc_mapping={
                                    200: 'InventoryMonitoringSchedulerInformation'
                                }
                            )

            class _savedsearches(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                def ID(self, id: str):
                    return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                def get(self):
                    class Output(CloudApiOutputModel):
                        SavedSearchResponse: outagedetection_service.SavedSearchResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'SavedSearchResponse'
                        }
                    )

                def post(self, SavedSearchRequest: outagedetection_service.SavedSearchRequest):
                    data = {**SavedSearchRequest.dict()}

                    class Output(CloudApiOutputModel):
                        SavedSearchResponse: outagedetection_service.SavedSearchResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'SavedSearchResponse'
                        }
                    )

                class _ID(CloudApiEndpoint):
                    def delete(self):
                        class Output(CloudApiOutputModel):
                            pass

                        return generate_output(output_cls=Output, response=self._delete(params={}))

                    def get(self):
                        class Output(CloudApiOutputModel):
                            SavedSearchInfo: outagedetection_service.SavedSearchInfo

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'SavedSearchInfo'
                            }
                        )

                    def put(self, SavedSearchRequest: outagedetection_service.SavedSearchRequest):
                        data = {**SavedSearchRequest.dict()}

                        class Output(CloudApiOutputModel):
                            SavedSearchInfo: outagedetection_service.SavedSearchInfo

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'SavedSearchInfo'
                            }
                        )



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ProxyAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Proxy"
    bypass_proxy_on_local = Attribute('Bypass Proxy on Local', min_version='21.4')
    credential = Attribute('Credential', min_version='21.4')
    proxy_credential = Attribute('Proxy Credential', min_version='21.4')
    proxy_host = Attribute('Proxy Host', min_version='21.4')
    proxy_port = Attribute('Proxy Port', min_version='21.4')
    proxy_use_host_configuration = Attribute('Proxy Use Host Configuration', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.schedule_base import ScheduleBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class OnboardDiscoveryAttributes(ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Onboard Discovery"
    amazon_account_ids = Attribute('Amazon Account Ids')
    application_type = Attribute('Application Type')
    azure_application_id = Attribute('Azure Application Id')
    azure_tenant_id = Attribute('Azure Tenant Id')
    certificates_placement_folder = Attribute('Certificates Placement Folder')
    credential = Attribute('Credential')
    default_container = Attribute('Default Container')
    device = Attribute('Device')
    devices_folder = Attribute('Devices Folder')
    disable_installations_not_in_use = Attribute('Disable Installations Not In Use')
    driver_name = Attribute('Driver Name')
    environment = Attribute('Environment')
    exclude_binding_targets = Attribute('Exclude Binding Targets')
    extract_private_key = Attribute('Extract Private Key')
    in_progress = Attribute('In Progress')
    last_run = Attribute('Last Run')
    last_update = Attribute('Last Update')
    log_debug = Attribute('Log Debug')
    onboard_discovery_import_work_to_do = Attribute('Onboard Discovery Import Work To Do')
    port = Attribute('Port')
    profiles_to_import = Attribute('Profiles To Import')
    result_count = Attribute('Result Count')
    script_execution_timeout = Attribute('Script Execution Timeout')
    script_hash_mismatch_error = Attribute('Script Hash Mismatch Error')
    status = Attribute('Status')
    stop_requested = Attribute('Stop Requested')
    timeout = Attribute('Timeout')
    xml_port = Attribute('XML Port')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class WorkflowAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Workflow"
    rule = Attribute('Rule', min_version='21.4')
    rule_vault_id = Attribute('Rule Vault Id', min_version='21.4')



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    CloudApiEndpoint,
    CloudApiOutputModel,
    generate_output,
)
from pyvenafi.cloud.api.models import account_service
from typing import List
from uuid import UUID

class _account_service:
    def __init__(self, api_obj):
        self.v1 = self._v1(api_obj=api_obj)

    class _v1(CloudApiEndpoint):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='v1')
            self.companies = self._companies(api_obj=self._api_obj, url=f'{self._url}/companies')
            self.dataencryptionkeys = self._dataencryptionkeys(
                api_obj=self._api_obj,
                url=f'{self._url}/dataencryptionkeys'
            )
            self.notifications = self._notifications(api_obj=self._api_obj, url=f'{self._url}/notifications')
            self.preferences = self._preferences(api_obj=self._api_obj, url=f'{self._url}/preferences')
            self.ssoconfigurations = self._ssoconfigurations(
                api_obj=self._api_obj,
                url=f'{self._url}/ssoconfigurations'
            )
            self.teams = self._teams(api_obj=self._api_obj, url=f'{self._url}/teams')
            self.useraccounts = self._useraccounts(api_obj=self._api_obj, url=f'{self._url}/useraccounts')
            self.users = self._users(api_obj=self._api_obj, url=f'{self._url}/users')

        class _companies(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def URLPREFIX(self, urlprefix: str):
                return self._URLPREFIX(api_obj=self._api_obj, url=f'{self._url}/{urlprefix}')

            class _ID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.urlPrefix = self._urlPrefix(api_obj=self._api_obj, url=f'{self._url}/urlPrefix')

                def get(self):
                    class Output(CloudApiOutputModel):
                        CompanyInformation: account_service.CompanyInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'CompanyInformation'
                        }
                    )

                class _urlPrefix(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            CompanyInformation: account_service.CompanyInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CompanyInformation'
                            }
                        )

            class _URLPREFIX(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.baseenv = self._baseenv(api_obj=self._api_obj, url=f'{self._url}/baseenv')
                    self.loginconfig = self._loginconfig(api_obj=self._api_obj, url=f'{self._url}/loginconfig')

                class _baseenv(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            CompanyInformation: account_service.CompanyInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CompanyInformation'
                            }
                        )

                class _loginconfig(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            CompanyLoginConfigResponse: account_service.CompanyLoginConfigResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CompanyLoginConfigResponse'
                            }
                        )

        class _dataencryptionkeys(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.rotation = self._rotation(api_obj=self._api_obj, url=f'{self._url}/rotation')

            class _rotation(CloudApiEndpoint):
                def post(self):
                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._post(data={}))

        class _notifications(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.type = self._type(api_obj=self._api_obj, url=f'{self._url}/type')

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self):
                class Output(CloudApiOutputModel):
                    NotificationConfigurationResponse: account_service.NotificationConfigurationResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'NotificationConfigurationResponse'
                    }
                )

            def post(self, NotificationConfigurationRequest: account_service.NotificationConfigurationRequest):
                data = {**NotificationConfigurationRequest.dict()}

                class Output(CloudApiOutputModel):
                    NotificationConfigurationInformation: account_service.NotificationConfigurationInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'NotificationConfigurationInformation'
                    }
                )

            class _ID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.unsubscribe = self._unsubscribe(api_obj=self._api_obj, url=f'{self._url}/unsubscribe')

                def delete(self):
                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._delete(params={}))

                def get(self):
                    class Output(CloudApiOutputModel):
                        NotificationConfigurationInformation: account_service.NotificationConfigurationInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'NotificationConfigurationInformation'
                        }
                    )

                def put(self, NotificationConfigurationRequest: account_service.NotificationConfigurationRequest):
                    data = {**NotificationConfigurationRequest.dict()}

                    class Output(CloudApiOutputModel):
                        NotificationConfigurationInformation: account_service.NotificationConfigurationInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._put(data=data),
                        rc_mapping={
                            200: 'NotificationConfigurationInformation'
                        }
                    )

                class _unsubscribe(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)

                    def RECIPIENTTOKEN(self, recipienttoken: str):
                        return self._RECIPIENTTOKEN(api_obj=self._api_obj, url=f'{self._url}/{recipienttoken}')

                    class _RECIPIENTTOKEN(CloudApiEndpoint):
                        def put(self):
                            class Output(CloudApiOutputModel):
                                UnsubscribeNotificationInformation: account_service.UnsubscribeNotificationInformation

                            return generate_output(
                                output_cls=Output,
                                response=self._put(data={}),
                                rc_mapping={
                                    202: 'UnsubscribeNotificationInformation'
                                }
                            )

            class _type(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                def TYPE(self, type: str):
                    return self._TYPE(api_obj=self._api_obj, url=f'{self._url}/{type}')

                class _TYPE(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            NotificationConfigurationInformation: account_service.NotificationConfigurationInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'NotificationConfigurationInformation'
                            }
                        )

        class _preferences(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.name = self._name(api_obj=self._api_obj, url=f'{self._url}/name')

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self):
                class Output(CloudApiOutputModel):
                    UserPreferencesResponse: account_service.UserPreferencesResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'UserPreferencesResponse'
                    }
                )

            def post(self, UserPreferenceRequest: account_service.UserPreferenceRequest):
                data = {**UserPreferenceRequest.dict()}

                class Output(CloudApiOutputModel):
                    UserPreferenceInformation: account_service.UserPreferenceInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'UserPreferenceInformation'
                    }
                )

            class _ID(CloudApiEndpoint):
                def delete(self):
                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._delete(params={}))

                def get(self):
                    class Output(CloudApiOutputModel):
                        UserPreferenceInformation: account_service.UserPreferenceInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'UserPreferenceInformation'
                        }
                    )

                def put(self, UserPreferenceRequest: account_service.UserPreferenceRequest):
                    data = {**UserPreferenceRequest.dict()}

                    class Output(CloudApiOutputModel):
                        UserPreferenceInformation: account_service.UserPreferenceInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._put(data=data),
                        rc_mapping={
                            200: 'UserPreferenceInformation'
                        }
                    )

            class _name(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                def NAME(self, name: str):
                    return self._NAME(api_obj=self._api_obj, url=f'{self._url}/{name}')

                class _NAME(CloudApiEndpoint):
                    def get(self):
                        class Output(CloudApiOutputModel):
                            UserPreferenceInformation: account_service.UserPreferenceInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'UserPreferenceInformation'
                            }
                        )

        class _ssoconfigurations(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self):
                class Output(CloudApiOutputModel):
                    SsoConfigurationResponse: account_service.SsoConfigurationResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'SsoConfigurationResponse'
                    }
                )

            def post(self, SsoConfigurationRequest: account_service.SsoConfigurationRequest):
                data = {**SsoConfigurationRequest.dict()}

                class Output(CloudApiOutputModel):
                    SsoConfigurationInformation: account_service.SsoConfigurationInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'SsoConfigurationInformation'
                    }
                )

            class _ID(CloudApiEndpoint):
                def delete(self):
                    class Output(CloudApiOutputModel):
                        SsoConfigurationInformation: account_service.SsoConfigurationInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._delete(params={}),
                        rc_mapping={
                            204: 'SsoConfigurationInformation'
                        }
                    )

                def put(self, SsoConfigurationRequest: account_service.SsoConfigurationRequest):
                    data = {**SsoConfigurationRequest.dict()}

                    class Output(CloudApiOutputModel):
                        SsoConfigurationInformation: account_service.SsoConfigurationInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._put(data=data),
                        rc_mapping={
                            201: 'SsoConfigurationInformation'
                        }
                    )

        class _teams(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self):
                class Output(CloudApiOutputModel):
                    TeamsResponse: account_service.TeamsResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'TeamsResponse'
                    }
                )

            def post(self, CreateTeamRequest: account_service.CreateTeamRequest):
                data = {**CreateTeamRequest.dict()}

                class Output(CloudApiOutputModel):
                    TeamInformation: account_service.TeamInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'TeamInformation'
                    }
                )

            class _ID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.members = self._members(api_obj=self._api_obj, url=f'{self._url}/members')
                    self.owners = self._owners(api_obj=self._api_obj, url=f'{self._url}/owners')

                def delete(self):
                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._delete(params={}))

                def get(self):
                    class Output(CloudApiOutputModel):
                        TeamInformation: account_service.TeamInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'TeamInformation'
                        }
                    )

                def patch(self, UpdateTeamRequest: account_service.UpdateTeamRequest):
                    data = {**UpdateTeamRequest.dict()}

                    class Output(CloudApiOutputModel):
                        TeamInformation: account_service.TeamInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._patch(data=data),
                        rc_mapping={
                            200: 'TeamInformation'
                        }
                    )

                class _members(CloudApiEndpoint):
                    def delete(self, TeamMembersRequest: account_service.TeamMembersRequest):
                        data = {**TeamMembersRequest.dict()}

                        class Output(CloudApiOutputModel):
                            TeamInformation: account_service.TeamInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._delete(params=data),
                            rc_mapping={
                                200: 'TeamInformation'
                            }
                        )

                    def post(self, TeamMembersRequest: account_service.TeamMembersRequest):
                        data = {**TeamMembersRequest.dict()}

                        class Output(CloudApiOutputModel):
                            TeamInformation: account_service.TeamInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                200: 'TeamInformation'
                            }
                        )

                class _owners(CloudApiEndpoint):
                    def delete(self, TeamOwnersRequest: account_service.TeamOwnersRequest):
                        data = {**TeamOwnersRequest.dict()}

                        class Output(CloudApiOutputModel):
                            TeamInformation: account_service.TeamInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._delete(params=data),
                            rc_mapping={
                                200: 'TeamInformation'
                            }
                        )

                    def post(self, TeamOwnersRequest: account_service.TeamOwnersRequest):
                        data = {**TeamOwnersRequest.dict()}

                        class Output(CloudApiOutputModel):
                            TeamInformation: account_service.TeamInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                200: 'TeamInformation'
                            }
                        )

        class _useraccounts(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.activation = self._activation(api_obj=self._api_obj, url=f'{self._url}/activation')
                self.activationresend = self._activationresend(
                    api_obj=self._api_obj,
                    url=f'{self._url}/activationresend'
                )
                self.apikeyrotation = self._apikeyrotation(api_obj=self._api_obj, url=f'{self._url}/apikeyrotation')
                self.invitations = self._invitations(api_obj=self._api_obj, url=f'{self._url}/invitations')
                self.password = self._password(api_obj=self._api_obj, url=f'{self._url}/password')
                self.passwordreset = self._passwordreset(api_obj=self._api_obj, url=f'{self._url}/passwordreset')
                self.updatepassword = self._updatepassword(api_obj=self._api_obj, url=f'{self._url}/updatepassword')

            def get(self):
                class Output(CloudApiOutputModel):
                    UserAccountResponse: account_service.UserAccountResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'UserAccountResponse'
                    }
                )

            def post(self, UserAccountRequest: account_service.UserAccountRequest):
                data = {**UserAccountRequest.dict()}

                class Output(CloudApiOutputModel):
                    UserAccountResponse: account_service.UserAccountResponse

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'UserAccountResponse',
                        202: 'UserAccountResponse'
                    }
                )

            class _activation(CloudApiEndpoint):
                def get(self, k: UUID, v: bool):
                    data = {
                        'k': k,
                        'v': v,
                    }

                    class Output(CloudApiOutputModel):
                        UserInformation: account_service.UserInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'UserInformation'
                        }
                    )

            class _activationresend(CloudApiEndpoint):
                def post(self, urlPrefix: str):
                    data = {
                        'urlPrefix': urlPrefix,
                    }

                    class Output(CloudApiOutputModel):
                        ResendActivationResponse: account_service.ResendActivationResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            202: 'ResendActivationResponse'
                        }
                    )

            class _apikeyrotation(CloudApiEndpoint):
                def get(self, k: UUID, v: bool):
                    data = {
                        'k': k,
                        'v': v,
                    }

                    class Output(CloudApiOutputModel):
                        ApiKeyInformation: account_service.ApiKeyInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'ApiKeyInformation'
                        }
                    )

            class _invitations(CloudApiEndpoint):
                def get(self, id: UUID):
                    data = {
                        'id': id,
                    }

                    class Output(CloudApiOutputModel):
                        UserInformation: account_service.UserInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'UserInformation'
                        }
                    )

                def post(self, InvitationRequest: account_service.InvitationRequest):
                    data = {**InvitationRequest.dict()}

                    class Output(CloudApiOutputModel):
                        InvitationResponse: account_service.InvitationResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'InvitationResponse'
                        }
                    )

                def put(self, InvitationConfirmationRequest: account_service.InvitationConfirmationRequest):
                    data = {**InvitationConfirmationRequest.dict()}

                    class Output(CloudApiOutputModel):
                        UserInformation: account_service.UserInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._put(data=data),
                        rc_mapping={
                            202: 'UserInformation'
                        }
                    )

            class _password(CloudApiEndpoint):
                def post(self, ChangePasswordRequest: account_service.ChangePasswordRequest):
                    data = {**ChangePasswordRequest.dict()}

                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._post(data=data))

            class _passwordreset(CloudApiEndpoint):
                def get(self, token: str):
                    data = {
                        'token': token,
                    }

                    class Output(CloudApiOutputModel):
                        ApiKeyInformation: account_service.ApiKeyInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            201: 'ApiKeyInformation'
                        }
                    )

                def post(self, urlPrefix: str):
                    data = {
                        'urlPrefix': urlPrefix,
                    }

                    class Output(CloudApiOutputModel):
                        ResetPasswordResponse: account_service.ResetPasswordResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            202: 'ResetPasswordResponse'
                        }
                    )

            class _updatepassword(CloudApiEndpoint):
                def post(self, UpdatePasswordRequest: account_service.UpdatePasswordRequest):
                    data = {**UpdatePasswordRequest.dict()}

                    class Output(CloudApiOutputModel):
                        UpdatePasswordResponse: account_service.UpdatePasswordResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'UpdatePasswordResponse'
                        }
                    )

        class _users(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.username = self._username(api_obj=self._api_obj, url=f'{self._url}/username')

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def USERID(self, userid: str):
                return self._USERID(api_obj=self._api_obj, url=f'{self._url}/{userid}')

            def get(self, userStatus: List[account_service.UserStatus], username: str):
                data = {
                    'userStatus': userStatus,
                    'username'  : username,
                }

                class Output(CloudApiOutputModel):
                    UserResponse: account_service.UserResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params=data),
                    rc_mapping={
                        200: 'UserResponse'
                    }
                )

            class _ID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.accounttype = self._accounttype(api_obj=self._api_obj, url=f'{self._url}/accounttype')
                    self.disabled = self._disabled(api_obj=self._api_obj, url=f'{self._url}/disabled')
                    self.locallogin = self._locallogin(api_obj=self._api_obj, url=f'{self._url}/locallogin')
                    self.roles = self._roles(api_obj=self._api_obj, url=f'{self._url}/roles')

                def get(self):
                    class Output(CloudApiOutputModel):
                        UserInformation: account_service.UserInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'UserInformation'
                        }
                    )

                class _accounttype(CloudApiEndpoint):
                    def put(self, UserAccountTypeRequest: account_service.UserAccountTypeRequest):
                        data = {**UserAccountTypeRequest.dict()}

                        class Output(CloudApiOutputModel):
                            UserInformation: account_service.UserInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'UserInformation'
                            }
                        )

                class _disabled(CloudApiEndpoint):
                    def put(self, UserDisabledRequest: account_service.UserDisabledRequest):
                        data = {**UserDisabledRequest.dict()}

                        class Output(CloudApiOutputModel):
                            UserInformation: account_service.UserInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'UserInformation'
                            }
                        )

                class _locallogin(CloudApiEndpoint):
                    def put(self, LocalLoginRequest: account_service.LocalLoginRequest):
                        data = {**LocalLoginRequest.dict()}

                        class Output(CloudApiOutputModel):
                            UserInformation: account_service.UserInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'UserInformation'
                            }
                        )

                class _roles(CloudApiEndpoint):
                    def put(self, RolesRequest: account_service.RolesRequest):
                        data = {**RolesRequest.dict()}

                        class Output(CloudApiOutputModel):
                            UserInformation: account_service.UserInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._put(data=data),
                            rc_mapping={
                                200: 'UserInformation'
                            }
                        )

            class _USERID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.apikeys = self._apikeys(api_obj=self._api_obj, url=f'{self._url}/apikeys')

                class _apikeys(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)

                    def KEY(self, key: str):
                        return self._KEY(api_obj=self._api_obj, url=f'{self._url}/{key}')

                    def get(self, apiKeyStatus: List[account_service.ApiKeyStatus]):
                        data = {
                            'apiKeyStatus': apiKeyStatus,
                        }

                        class Output(CloudApiOutputModel):
                            ApiKeyResponse: account_service.ApiKeyResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params=data),
                            rc_mapping={
                                200: 'ApiKeyResponse'
                            }
                        )

                    def post(self, ApiKeyRequest: account_service.ApiKeyRequest):
                        data = {**ApiKeyRequest.dict()}

                        class Output(CloudApiOutputModel):
                            ApiKeyResponse: account_service.ApiKeyResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                201: 'ApiKeyResponse'
                            }
                        )

                    class _KEY(CloudApiEndpoint):
                        def __init__(self, *args, **kwargs):
                            super().__init__(*args, **kwargs)
                            self.replacement = self._replacement(api_obj=self._api_obj, url=f'{self._url}/replacement')
                            self.rotation = self._rotation(api_obj=self._api_obj, url=f'{self._url}/rotation')
                            self.rotationrequest = self._rotationrequest(
                                api_obj=self._api_obj,
                                url=f'{self._url}/rotationrequest'
                            )

                        def get(self):
                            class Output(CloudApiOutputModel):
                                ApiKeyInformation: account_service.ApiKeyInformation

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params={}),
                                rc_mapping={
                                    200: 'ApiKeyInformation'
                                }
                            )

                        class _replacement(CloudApiEndpoint):
                            def post(self, ApiKeyRequest: account_service.ApiKeyRequest):
                                data = {**ApiKeyRequest.dict()}

                                class Output(CloudApiOutputModel):
                                    ApiKeyResponse: account_service.ApiKeyResponse

                                return generate_output(
                                    output_cls=Output,
                                    response=self._post(data=data),
                                    rc_mapping={
                                        200: 'ApiKeyResponse'
                                    }
                                )

                        class _rotation(CloudApiEndpoint):
                            def put(self):
                                class Output(CloudApiOutputModel):
                                    ApiKeyNullResponse: account_service.ApiKeyNullResponse

                                return generate_output(
                                    output_cls=Output,
                                    response=self._put(data={}),
                                    rc_mapping={
                                        200: 'ApiKeyNullResponse'
                                    }
                                )

                        class _rotationrequest(CloudApiEndpoint):
                            def post(self, ApiKeyRequest: account_service.ApiKeyRequest):
                                data = {**ApiKeyRequest.dict()}

                                class Output(CloudApiOutputModel):
                                    ApiKeyResponse: account_service.ApiKeyResponse

                                return generate_output(
                                    output_cls=Output,
                                    response=self._post(data=data),
                                    rc_mapping={
                                        201: 'ApiKeyResponse'
                                    }
                                )

            class _username(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                def USERNAME(self, username: str):
                    return self._USERNAME(api_obj=self._api_obj, url=f'{self._url}/{username}')

                class _USERNAME(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                        self.loginconfig = self._loginconfig(api_obj=self._api_obj, url=f'{self._url}/loginconfig')

                    def get(self):
                        class Output(CloudApiOutputModel):
                            UserResponse: account_service.UserResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'UserResponse'
                            }
                        )

                    class _loginconfig(CloudApiEndpoint):
                        def get(self):
                            class Output(CloudApiOutputModel):
                                UserLoginConfigResponse: account_service.UserLoginConfigResponse

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params={}),
                                rc_mapping={
                                    200: 'UserLoginConfigResponse'
                                }
                            )



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class AgentCertificateBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Agent Certificate Base"
    certificate_scanner_capi = Attribute('Certificate Scanner CAPI', min_version='21.4')
    certificate_scanner_map = Attribute('Certificate Scanner Map', min_version='21.4')
    certificate_scanner_path = Attribute('Certificate Scanner Path', min_version='21.4')
    credential = Attribute('Credential', min_version='21.4')
    digest = Attribute('Digest', min_version='21.4')
    exclude_remote_mount_points = Attribute('Exclude Remote Mount Points', min_version='21.4')
    expiration = Attribute('Expiration', min_version='21.4')
    key_store_validation_disabled = Attribute('Key Store Validation Disabled', min_version='21.4')
    log_threshold = Attribute('Log Threshold', min_version='21.4')
    max_filesize = Attribute('Max Filesize', min_version='21.4')
    maximum_threads = Attribute('Maximum Threads', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.report_base import ReportBaseAttributes
from pyvenafi.tpp.attributes.report_filter_base import ReportFilterBaseAttributes

class ReportSSHAuthorizedUserAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Report:SSH Authorized User"
    options = Attribute('Options', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.workflow import WorkflowAttributes

class AdaptableWorkflowAttributes(WorkflowAttributes, metaclass=IterableMeta):
    __config_class__ = "Adaptable Workflow"
    adaptable_workflow_text_field_1 = Attribute('Adaptable Workflow Text Field 1', min_version='21.4')
    adaptable_workflow_text_field_10 = Attribute('Adaptable Workflow Text Field 10', min_version='21.4')
    adaptable_workflow_text_field_11 = Attribute('Adaptable Workflow Text Field 11', min_version='21.4')
    adaptable_workflow_text_field_12 = Attribute('Adaptable Workflow Text Field 12', min_version='21.4')
    adaptable_workflow_text_field_2 = Attribute('Adaptable Workflow Text Field 2', min_version='21.4')
    adaptable_workflow_text_field_3 = Attribute('Adaptable Workflow Text Field 3', min_version='21.4')
    adaptable_workflow_text_field_4 = Attribute('Adaptable Workflow Text Field 4', min_version='21.4')
    adaptable_workflow_text_field_5 = Attribute('Adaptable Workflow Text Field 5', min_version='21.4')
    adaptable_workflow_text_field_6 = Attribute('Adaptable Workflow Text Field 6', min_version='21.4')
    adaptable_workflow_text_field_7 = Attribute('Adaptable Workflow Text Field 7', min_version='21.4')
    adaptable_workflow_text_field_8 = Attribute('Adaptable Workflow Text Field 8', min_version='21.4')
    adaptable_workflow_text_field_9 = Attribute('Adaptable Workflow Text Field 9', min_version='21.4')
    credential = Attribute('Credential', min_version='21.4')
    log_debug = Attribute('Log Debug', min_version='21.4')
    oauth_token_application_id = Attribute('OAuth Token Application Id', min_version='21.4')
    oauth_token_credential = Attribute('OAuth Token Credential', min_version='21.4')
    oauth_token_scope = Attribute('OAuth Token Scope', min_version='21.4')
    powershell_script = Attribute('PowerShell Script', min_version='21.4')
    powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='21.4')
    script_execution_timeout = Attribute('Script Execution Timeout', min_version='21.4')
    secondary_credential = Attribute('Secondary Credential', min_version='21.4')
    service_address = Attribute('Service Address', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.ssh_key import SSHKeyAttributes
from pyvenafi.tpp.attributes.validation_base import ValidationBaseAttributes

class SSHServerKeyAttributes(SSHKeyAttributes, ValidationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "SSH Server Key"
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
